import os
import traceback
import json
from abc import ABC, abstractmethod

import pika

from .util import get_env, load_yaml_config
from .atlasLogger import AtlasLogger

# A tuple containing values which must be provided to use Heracles RabbitMQClient.
_REQUIRED_INIT_VALUES = (
    "RABBIT_MQ_USER",
    "RABBIT_MQ_PASS",
    "RABBIT_MQ_EXCHANGE_PRIMARY",
    "RABBIT_MQ_EXCHANGE_DEAD",
    "RABBIT_MQ_HOST",
    "RABBIT_MQ_PORT",
    "RABBIT_MQ_VIRTUAL_HOST",
    "BASE_DB_API",
)

_OPTIONAL_INIT_VALUES = ("REST_DB_API_KEY", "SENDGRID_API_MAIL_TOKEN")


class RabbitMQClient(ABC):
    def __init__(self, service_name, path):
        load_yaml_config(path=path)

        self.header_type = None
        self.environment = get_env()
        self.credentials = pika.PlainCredentials(
            os.getenv("RABBIT_MQ_USER"), os.getenv("RABBIT_MQ_PASS")
        )
        self.parameters = pika.ConnectionParameters(
            os.getenv("RABBIT_MQ_HOST"),
            os.getenv("RABBIT_MQ_PORT"),
            os.getenv("RABBIT_MQ_VIRTUAL_HOST"),
            self.credentials,
        )
        self.connection = pika.BlockingConnection(self.parameters)
        self.channel = self.connection.channel()
        self.counter = 0
        self.method = None
        self.delivery_tag = None
        self.exchange = os.getenv("RABBIT_MQ_EXCHANGE_PRIMARY")
        self.message_body = None
        self.headers = None
        self.logger = AtlasLogger(service_name)

    def open_connection(self):
        """
        Opening a connection. This method is used to reconnect if the connection has been broken/ terminated.
        :return: void
        """
        self.parameters = pika.ConnectionParameters(
            os.getenv("RABBIT_MQ_HOST"),
            os.getenv("RABBIT_MQ_PORT"),
            os.getenv("RABBIT_MQ_VIRTUAL_HOST"),
            self.credentials,
        )
        self.connection = pika.BlockingConnection(self.parameters)
        self.channel = self.connection.channel()

    def send_to_queue(self, body, headers, exchange=None):
        """
        Allow resources to send data to the exchange
        :param body: json
        :param headers: json
        :param exchange: optional string
        :param unique: optional string (if unique is not defined key 'unique' must be present in body)
        :return: void
        """
        routing_key = self.environment

        if not exchange:
            exchange = self.exchange

        if not isinstance(body, dict):
            self.failed_request(self, body=body)
            raise TypeError(f"Expected body type dict, instead got {type(body)}")

        body = json.dumps(body)

        try:
            self.channel.basic_publish(
                exchange=exchange,
                properties=pika.BasicProperties(headers=headers, delivery_mode=2),
                routing_key=routing_key,
                body=body,
            )
            self.counter = 0
            self.logger.debug(f"{body}")
        except Exception:
            traceback.print_exc()
            self.logger.error(
                message="failed to send to queue",
                message_body=body,
                error=traceback.format_exc(),
            )
            if self.counter < 3:
                self.open_connection()
                self.counter += 1
                self.send_to_queue(body=body, headers=headers, exchange=exchange)
            else:
                self.failed_request(self, body=body)

    def failed_request(self, body, method=None, log=True, *args, **kwargs):
        """
        When a request failed. This can be because the format or data was incorrect.
        :param log:
        :param method: object
        :param body: json
        :return: void
        """
        if log:
            self.logger.error(
                message="request failed, do something with it", message_body=body
            )

        # Check origin of request
        if method:
            # This request originated from listening to a queue,
            # therefore this message will have to be acknowledged to remove it from the queue
            self.channel.basic_ack(delivery_tag=method.delivery_tag)

        self.dead_message()

    @staticmethod
    def exists(data, error, search):
        """
        This function will check if a certain search term is available in the data
        :param data:
        :param error:
        :param search:
        :return: string
        """
        if data.get(search, None) is None:
            error[search] = f"{search} was not provided"
        return error

    @abstractmethod
    def validate(self):
        pass

    @abstractmethod
    def message_action(self):
        pass

    def ack_message(self, success=None):
        """Acknowledges message in RabbitMQ"""

        if self.channel.is_open:
            if success is not None:
                if success:
                    self.logger.info(
                        message=f"Acknowledged message with state: Success",
                        message_body=self.message_body,
                    )
                else:
                    self.logger.info(
                        message=f"Acknowledged message with state: Failed",
                        message_body=self.message_body,
                    )
            try:
                self.channel.basic_ack(self.delivery_tag)
            except Exception:
                self.channel.basic_ack(0)
        else:
            self.open_connection()
            self.ack_message(success=success)

    def dead_message(self):
        self.send_to_queue(
            body=self.message_body,
            headers=self.headers,
            exchange=os.getenv("RABBIT_MQ_EXCHANGE_DEAD"),
        )
        self.ack_message()

    def callback(self, ch, method, properties, body, *args, **kwargs):
        """callback for the message"""
        self.headers = properties.headers

        # Check if the header has a type defined. This is required for all remaining services
        if "type" in self.headers.keys():
            self.method = method
            try:
                self.message_body = json.loads(body.decode("utf-8"))
            except json.decoder.JSONDecodeError:
                self.message_body = {"NOT_JSON": f"{self.message_body}"}
                self.headers["type"] = "NOT JSON DONT USE"
            self.delivery_tag = method.delivery_tag

            if self.headers["type"] == self.header_type:
                self.logger.debug(
                    message="Received a new message", message_body=self.message_body,
                )
                if self.validate():
                    self.message_action()
                else:
                    self.dead_message()
            else:
                # Message was not meant for this service
                self.ack_message()
        else:
            # KeyError: Type is not defined in headers
            self.logger.warning(
                message="Message without type headers has been found in the queue, setting header to type:undefined.header",
                message_body=self.message_body,
            )
            self.headers["type"] = "undefined.header"
            self.callback(body=body, ch=ch, properties=properties, method=method)

    def start_receive(self, queue, header_type, consumer_tag=None):
        """Starts receiving the messages from a certain RabbitMQ queue"""
        self.header_type = header_type
        self.channel.basic_qos()
        self.channel.basic_consume(
            queue=queue,
            on_message_callback=self.callback,
            auto_ack=False,
            consumer_tag=self.environment,
        )
        self.channel.start_consuming()
