import socket
import os

import requests
import yaml
from .herculesExceptions import DatabaseException

_REQUIRED_INIT_VALUES = (
    "RABBIT_MQ_USER",
    "RABBIT_MQ_PASS",
    "RABBIT_MQ_EXCHANGE_PRIMARY",
    "RABBIT_MQ_EXCHANGE_DEAD",
    "RABBIT_MQ_HOST",
    "RABBIT_MQ_PORT",
    "RABBIT_MQ_VIRTUAL_HOST",
    "BASE_DB_API_URL",
)


def get_env():
    env = os.getenv("GITLAB_ENVIRONMENT_NAME")
    if env:
        if "review" in env:
            environment = f"{str(env)} - {socket.gethostname()}"
        elif "staging" in env:
            environment = f"staging-{socket.gethostname()}"
        elif "production" in env:
            environment = f"production-{socket.gethostname()}"
        else:
            environment = f"review-{socket.gethostname()}"
    else:
        environment = f"review-no-env-{socket.gethostname()}"
    return environment


def get_db(table, specific=None, limit=None, order_by=None, **kwargs):
    for key in kwargs:
        if not isinstance(kwargs[key], dict):
            kwargs[key] = {"_eq": kwargs[key]}
    if order_by:
        if not isinstance(order_by, dict):
            kwargs["order_by"] = {order_by: "desc"}
        else:
            kwargs["order_by"] = order_by
    if limit:
        kwargs["limit"] = limit
    response = requests.get(
        url=f"{os.getenv('BASE_DB_API_URL')}/{table}/v1/", json=kwargs
    ).json()
    if response.get("errors"):
        raise DatabaseException(error=str(response.get("errors")))
    if specific:
        if response.get("data").get(table):
            return response.get("data").get(table)[0].get(specific)
        return response.get("data").get(table)
    return response["data"][table]


def post_db(table, **kwargs):
    response = requests.post(
        url=f"{os.getenv('BASE_DB_API_URL')}/{table}/v1/", json=kwargs
    ).json()
    if response.get("errors"):
        raise DatabaseException(error=str(response.get("errors")))
    return response["data"][f"insert_{table}"]["returning"][0]


def put_db(table, put, **kwargs):
    if not isinstance(put, dict):
        raise TypeError(f"Put data is type {type(put)} not dict")
    for key in kwargs:
        if not isinstance(kwargs[key], dict):
            kwargs[key] = {"_eq": kwargs[key]}
    query = {"data": put, "where": kwargs}
    response = requests.put(
        url=f"{os.getenv('BASE_DB_API_URL')}/{table}/v1/", json=query
    ).json()
    if response.get("errors"):
        raise DatabaseException(error=str(response.get("errors")))
    if response["data"][f"update_{table}"]["returning"]:
        return response["data"][f"update_{table}"]["returning"][0]
    return response["data"][f"update_{table}"]["returning"]


def load_yaml_config(path):
    """Load RabbitMQClient information from yaml file
    Args:
      [optional]
      path: The path string to the file containing cached RabbitMQClient data.
    Raises:
      An IOError for the missing yaml file.
      A KeyError for missing keys in the yaml file.
    """
    with open(path, "rb") as handle:
        yaml_doc = handle.read()
    product_data = yaml.safe_load(yaml_doc)

    client_kwargs = {}
    try:
        for key in _REQUIRED_INIT_VALUES:
            client_kwargs[key] = product_data["RABBIT_MQ"][key]
    except KeyError:
        raise KeyError(
            f"Required values are missing in config.yaml. Required values are {_REQUIRED_INIT_VALUES}"
        )

    for key, value in product_data["OPTIONAL"].items():
        client_kwargs[key] = value

    for key, value in client_kwargs.items():
        os.environ[key] = value
